#
#  $Id: dtkmain.py,v 1.8 2001/12/25 06:55:41 rob Exp $
#
#  Copyright 2001 Rob Tillotson <rob@pyrite.org>
#  All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee or royalty is
#  hereby granted, provided that the above copyright notice appear in
#  all copies and that both the copyright notice and this permission
#  notice appear in supporting documentation or portions thereof,
#  including modifications, that you you make.
#
#  THE AUTHOR ROB TILLOTSON DISCLAIMS ALL WARRANTIES WITH REGARD TO
#  THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
#  AND FITNESS.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
#  SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
#  RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
#  CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
#  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE!
#
"""
"""

__version__ = '$Id: dtkmain.py,v 1.8 2001/12/25 06:55:41 rob Exp $'

__copyright__ = 'Copyright 2001 Rob Tillotson <rob@pyrite.org>'

import sys, os, string
from pprint import pprint

VERSION = '1.99.4'
DATE = "2001/12/20"
IDENT = "Pyrite Publisher v%s (%s) by Rob Tillotson <rob@pyrite.org>" % (VERSION, DATE)

import plugin, dtkplugins, config


_option = plugin.CLIOptionDescriptor
main_options = [
    _option('help', 'h', 'help', 'Get help', boolean=1),
    _option('list_plugins', 'l', 'list-plugins', 'List available formats', boolean=1),
    _option('force_plugins', 'P', None, 'Force specific plugins', vtype='NAME', multiple=1),
    _option('list_properties', None, 'list-properties', 'List plugin properties', boolean=1),
    ]


class ConversionError(Exception): pass

class Placeholder: pass


# The way plugin autoselection works is more or less as follows:
#
#   - input plugins are chosen based on their ability to handle the
#     filename presented to them (normally this should be done
#     automatically)
#
#     the input plugin, after being given the filename, will return
#     the name of a mimetype for the file.  (This is, in effect, an
#     interface name, for purposes of the chaining described below.)
#
#   - parsers and assemblers are chained according to pairs of
#     interfaces.  They can report to the main program which pairs
#     of (input, output) they can handle, and a priority level
#     which can be used to select an interface when there are more
#     than one choice.
#
#   - output plugins can report which input interfaces they can
#     handle.
#
# To autoselect, first an input plugin is selected, and the mimetype
# of the file determined.  Then, the program will determine all
# possible chains of parser/assembler/output plugins from that mimetype.
# If there is more than one it will pick the "best" one.
#

# Since this architecture supports a variable chain of plugins (except
# for the input, and the fact that it must end with an output) the
# method of "stacking" them will now be more uniform...
#
# Plugins will communicate via "events" which are just method calls on
# an object.  Calls will propagate from input to output; the "driver"
# will be wedged between the input and the first plugin, feeding data
# into the first plugin and letting everything cascade from there.
#
# The way this will work is that at the beginning of processing, each
# plugin will have its "open" method called with the chain as argument,
# and this should return an API object which will be handed to the
# *previous* plugin in the chain once all plugins are open.

class Chain:
    def __init__(self, plugins):
	self.plugins = plugins

	self.collect_path_info()

	self.path = []

    def debug(self):
	pprint(self.__f_links)
	print
	
    def pathfind(self, start, end):
	# f_links and r_links are the link dicts built by the caller
	# paths is the list of paths known so far -- it is a dict
	# and we exploit python's referenceness to modify it
	# and also to not have to recurse TOO much because we can
	# look in it to find paths that have already been calculated

	# returns a list of all known paths from start to end
	# and populates the paths parameter as a cache

	f_links = self.__f_links
	
	paths = []
	if not f_links.has_key(start):
	    return []
	
	for c in f_links[start].keys():  # all possible "next steps" from start
	    if c == end:
		for pri, plugin in f_links[start][c]:
		    paths.append( (pri, (plugin,)) )
	    else:
		tailpaths = self.pathfind(c, end)
		# take all the returned paths, and prepend all of the plugins
		# which could handle them
		for pathpri, path in tailpaths:
		    for pri, plugin in f_links[start][c]:
			newpri = (pathpri*len(path)+pri)/(len(path)+1)
			newpath = (plugin,)+path
			paths.append( (newpri, newpath) )
			
	return paths
    
		    
	
    def collect_path_info(self):
	"""Collects information about possible chains from plugins.
	Every plugin must have a 'get_supported_links' function which
	returns a list of valid chainlinks for this plugin, in the form:
	  [ ( input, output, priority), ... ]
	Note that input plugins are special, because they can return different
	mimetypes, but all other plugins use this interface.  Also note that
	the link specifications must be tuples, because they will be used
	as dictionary keys.

	If the output parameter is None this is an output plugin.
	"""
	links = {}
	r_links = {}
	for plugin in self.plugins:
	    lks = plugin.get_supported_links()
	    for pri, inp, outp in lks:
		if not links.has_key(inp): links[inp] = {}
		if not links[inp].has_key(outp): links[inp][outp] = []
		links[inp][outp].append((pri, plugin))

		if not r_links.has_key(outp): r_links[outp] = {}
		if not r_links[outp].has_key(inp): r_links[outp][inp] = []
		r_links[outp][inp].append((pri, plugin))

	    
	self.__f_links = links
	self.__r_links = r_links


    def open(self, plugins, *a, **kw):
	self.path = plugins[:]
	path = plugins[:]
	path.reverse()

	last = None
	for p in path:
	    last = apply(p.open, (self, last)+a, kw)

	return last

    def close(self):
	for p in self.path:
	    if hasattr(p, 'close') and callable(p.close):
		p.close()
	self.path = []
	    

    def check_plugin(self, namelist, plugin, meth="handles_interface"):
	"""Tests whether a plugin supports one of the named interfaces.
	"""
	l = filter(lambda x, p=plugin, m=meth: getattr(p,m) is not None,
		   map(lambda x: x[1], namelist))
	if not l:
	    raise ConversionError, ("plugin '%s' does not support any of " + \
				    string.join(map(lambda x: x[1], namelist), ',')) \
				    % repr(plugin)
	else:
	    return l[0]
	

    def choose_best_plugin(self, namelist, plugins, meth="handles_interface"):
	"""Choose the best plugin to handle one of several named interfaces.
	"""
	nl = namelist[:]
	nl.sort()
	nl.reverse()
	for mpri, mtype in namelist:
	    l = map(lambda x, t=mtype, m=meth: (getattr(x,m)(t), x), plugins)
	    l.sort()
	    if l:
		return (l[-1][1], mtype)

	raise ConversionError, ("can't find a match for " +
				string.join(map(lambda x: x[1], namelist), ','))
    
	    
	

def main(argv):
    progname = argv[0]
    argv = argv[1:]

    # what a hack :)
    if "--DEBUG" in argv:
        ll = 3
        argv.remove("--DEBUG")
    else:
        ll = 1
        
    callback = dtkplugins.DTKCallback(loglevel=ll)

    plugins = dtkplugins.load_all_plugins(callback)

    conf = config.PPConfigProcessor(plugins)
    conf.load(os.path.expanduser('~/.pyrpubrc'))

    options = Placeholder()
    
    oparser = plugin.CLIOptionParser(main_options, options, plugins.values())

    options.help = 0
    options.list_plugins = 0
    options.list_properties = 0
    options.force_plugins = []
    argv = oparser(argv) # no longer need partial=1

    # -P option
    forced_plugins = []
    if options.force_plugins:
        for arg in options.force_plugins:
            for pname in map(string.strip, string.split(arg,',')):
                if not plugins.has_key(pname):
                    print plugins.keys()
                    print "error: no such plugin '%s'" % pname
                else:
                    forced_plugins.append(plugins[pname])
		    

    #oparser = plugin.CLIOptionParser([], None, [in_plug, iplug, oplug, out_plug])
    #argv = oparser(argv)

    # first, see if we are doing a -l, if so just do it and quit
    
    if options.list_plugins:
	ky = plugins.keys()
	ky.sort()
        print IDENT
        print
        print "Plugin           Description"
        print "------           -----------"
	for k in ky:
	    p = plugins[k]
	    print "%-15s  %-59s" % (p.name, string.split(p.description, '\n')[0][:60])
	return

    if options.help:
	print IDENT
	print
	print oparser.help(main_options)
	print

	ky = plugins.keys()
	ky.sort()
	for k in ky:
	    ip = plugins[k]
	    if ip.getCLIOptionDescriptors():
		print "Plugin: %s %s by %s <%s>" % (ip.name, ip.version, ip.author,
						    ip.author_email)
		if ip.description: print string.split(ip.description,'\n')[0]

		print
		print oparser.help(ip.getCLIOptionDescriptors())
		print
        return
    
    if options.list_properties:
        print IDENT
        print
        print "! = boolean property, # = multiple values allowed"
        print
        ky = plugins.keys()
        ky.sort()
        for k in ky:
            ip = plugins[k]
            props = filter(lambda x: not x.read_only, ip.getPropertyDescriptors())
            if props:
                props.sort()
                print "Plugin: %s %s by %s <%s>" % (ip.name, ip.version, ip.author,
                                                    ip.author_email)
                print
                for prop in props:
                    if prop.indexed: ty = '#'
                    elif prop.boolean: ty = '!'
                    else: ty = ' '
                    d = string.split(prop.description,'\n')[0][:54]
                    print "  %-20s %s %s" % (prop.name, ty, d)
                print
        return
    
    # *********************************************
    
    chain = Chain(plugins.values())

    for fn in argv:
	print "Converting %s..." % fn,
	sys.stdout.flush()

	try:
	    # first, pick an input plugin
	
	    l = filter(lambda x: x is not None and x[0] is not None,
		       map(lambda x, n=fn:
			   hasattr(x,"handles_filename")
			   and (x.handles_filename(n),x) or None,
			   plugins.values()))
	    l.sort()
	    if not l:
		raise ConversionError, "no plugin knows how to handle '%s'" % fn

	    iplug = l[-1][1]
	    mimetypes, basename = iplug.open_input(fn)

	    # the input plugin can return a list of mimetypes, which will be
	    # tried in order to find a path.
	    paths = None
	    for mimetype in mimetypes:
		paths = chain.pathfind(mimetype, None)
		# filter out paths which don't contain forced plugins
		for pi in forced_plugins:
		    paths = filter(lambda x,p=pi: p in x[1], paths)
		
		if paths: break

	    if not paths:
                mt = mimetypes[:]
                if len(mt) > 1: mt.insert(len(mt)-1, 'or')
		raise ConversionError, "couldn't find a way to convert from %s" % string.join(mt)
	    
	    # DEBUG
	    #print "Paths after forcing plugins:"
	    #for p in paths:
		#print "%3d:" % p[0],
		#for pi in p[1]: print pi.name, "->",
		#print
	    # /DEBUG
	    
	    paths.sort()
	    path = list(paths[-1][1])

	    # DEBUG
	    print "("+string.join(map(lambda x: x.name, (iplug,)+tuple(path)),',')+")",
	    sys.stdout.flush()
	    # /DEBUG
	    
	    head = chain.open(path, basename)
	    input = iplug.open(chain, head)

	    input.go(mimetype)

	    chain.close()
	    input.close_input()

	    print "done."
            
	except ConversionError, err:
	    print "error: %s" % err
	    continue

	except IOError, err:
	    print "i/o error: %s" % err
	    continue
	
